#!/usr/bin/env python
import collada
import numpy

import ctypes

from OpenGL.GL import *;
from OpenGL.GLU import *;

import glutils
from glutils import VecF
import shader
from shader import Shader
import shaders


class GLSLRenderer: 

    def __init__(self, dae):
        self.dae = dae
        # To calculate model boundary along Z axis
        self.z_max = -100000.0
        self.z_min = 100000.0
        self.textures = {}
        self.shaders = {}
        self.batch_list = []
        
        self.gllist = None;

        # Initialize OpenGL
        #glClearColor(0.0, 0.0, 0.0, 0.5) # Black Background
        glEnable(GL_DEPTH_TEST) # Enables Depth Testing
        glEnable(GL_CULL_FACE)
        glEnable(GL_MULTISAMPLE);

        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glEnable(GL_LIGHTING)

        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, VecF(0.1, 0.1, 0.1, 1.0))
        glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, VecF(0.1, 0.1, 0.1, 1.0))
        glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 50)

        #(vert, frag) = shaders.ADSPhong
        (vert, frag) = shaders.simplePhong
        prog = Shader(vert, frag)
        print '  phong'
        self.shaders['phong'] = prog
        (vert, frag) = shaders.pointLightDiff
        prog = Shader(vert, frag)
        self.shaders['lambert'] = prog
        print '  lambert'
        self.shaders['blinn'] = prog
        print '  blinn'
        (vert, frag) = shaders.flatShader
        prog = Shader(vert, frag)
        self.shaders['constant'] = prog
        print '  constant'
        (vert, frag) = shaders.texturePhong
        prog = Shader(vert, frag)
        self.shaders['texture'] = prog

        if self.dae.scene is not None:
            for geom in self.dae.scene.objects('geometry'):
                for prim in geom.primitives():
                    mat = prim.material
                    diff_color = VecF(1.0,1.0,1.0,1.0)
                    #diff_color = None;
                    spec_color = None 
                    shininess = None
                    amb_color = None
                    tex_id = None
                    shader_prog = self.shaders[mat.effect.shadingtype]
                    #shader_prog = None;
                    for prop in mat.effect.supported:
                        value = getattr(mat.effect, prop)
                        print mat.effect;
                        print value;
                        # it can be a float, a color (tuple) or a Map
                        # ( a texture )
                        if isinstance(value, collada.material.Map):
                            colladaimage = value.sampler.surface.image

                            # Accessing this attribute forces the
                            # loading of the image using PIL if
                            # available. Unless it is already loaded.
                            img = colladaimage.pilimage
                            if img: # can read and PIL available
                                shader_prog = self.shaders['texture']
                                #shader_prog = None;
                                # See if we already have texture for this image
                                if self.textures.has_key(colladaimage.id):
                                    tex_id = self.textures[colladaimage.id]
                                else:
                                    # If not - create new texture
                                    try:
                                        # get image meta-data
                                        # (dimensions) and data
                                        (ix, iy, tex_data) = (img.size[0], img.size[1], img.tostring("raw", "RGBA", 0, -1))
                                    except SystemError:
                                        # has no alpha channel,
                                        # synthesize one
                                        (ix, iy, tex_data) = (img.size[0], img.size[1], img.tostring("raw", "RGBX", 0, -1))
                                    # generate a texture ID
                                    tid = GLuint()
                                    glGenTextures(1, ctypes.byref(tid))
                                    tex_id = tid.value
                                    # make it current
                                    glBindTexture(GL_TEXTURE_2D, tex_id)
                                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
                                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
                                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
                                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
                                    # copy the texture into the
                                    # current texture ID
                                    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ix, iy, 0, GL_RGBA, GL_UNSIGNED_BYTE, tex_data)

                                    self.textures[colladaimage.id] = tex_id
                            else:
                                print '  %s = Texture %s: (not available)'%(
                                    prop, colladaimage.id)
                        else:
                            if prop == 'diffuse' and value is not None:
                                diff_color = value
                            elif prop == 'specular' and value is not None:
                                spec_color = value
                            elif prop == 'ambient' and value is not None:
                                amb_color = value
                            elif prop == 'shininess' and value is not None:
                                shininess = value

                    # use primitive-specific ways to get triangles
                    prim_type = type(prim).__name__
                    if prim_type == 'BoundTriangleSet':
                        triangles = prim
                    elif prim_type == 'BoundPolylist':
                        triangles = prim.triangleset()
                    else:
                        print 'Unsupported mesh used:', prim_type
                        triangles = None

                    if triangles is not None:
                        triangles.generateNormals()
                        # We will need flat lists for VBO (batch) initialization
                        vertices = triangles.vertex.flatten().tolist()
                        batch_len = len(vertices)//3
                        indices = triangles.vertex_index.flatten().tolist()
                        normals = triangles.normal.flatten().tolist()

                        #batch = pyglet.graphics.Batch()

                        # Track maximum and minimum Z coordinates
                        # (every third element) in the flattened
                        # vertex list
                        ma = max(vertices[2::3])
                        if ma > self.z_max:
                            self.z_max = ma

                        mi = min(vertices[2::3])
                        if mi < self.z_min:
                            self.z_min = mi

                        if tex_id is not None:
                            print "Texture...";

                            # This is probably the most inefficient
                            # way to get correct texture coordinate
                            # list (uv). I am sure that I just do not
                            # understand enough how texture
                            # coordinates and corresponding indexes
                            # are related to the vertices and vertex
                            # indicies here, but this is what I found
                            # to work. Feel free to improve the way
                            # texture coordinates (uv) are collected
                            # for batch.add_indexed() invocation.
                            uv = [[0.0,0.0]] * batch_len
                            for t in triangles:
                                nidx = 0
                                texcoords = t.texcoords[0]
                                for vidx in t.indices:
                                    uv[vidx] = texcoords[nidx].tolist()
                                    nidx += 1
                            # Flatten the uv list
                            uv = [item for sublist in uv for item in sublist]

                            # Create textured batch
                            #batch.add_indexed(batch_len, 
                            #                 GL_TRIANGLES,
                            #                 None,
                            #                 indices,
                            #                 ('v3f/static', vertices),
                            #                 ('n3f/static', normals),
                            #                 ('t2f/static', uv))
                            self.batch_list.append(
                                ([batch_len, GL_TRIANGLES, None, indices, ('v3f/static', vertices), ('n3f/static', normals), ('t2f/static', uv)], shader_prog, tex_id, diff_color, 
                                spec_color, amb_color, shininess))
                        else:
                            self.batch_list.append(
                                ([batch_len, GL_TRIANGLES, None, indices, ('v3f/static', vertices), ('n3f/static', normals)], shader_prog, tex_id, diff_color, 
                                spec_color, amb_color, shininess))
                            # Create colored batch
                            #batch.add_indexed(batch_len, 
                            #                 GL_TRIANGLES,
                            #                 None,
                            #                 indices,
                            #                 ('v3f/static', vertices),
                            #                 ('n3f/static', normals))

        print 'done. Ready to render.'

    def render(self, x, y, z, rotate_x, rotate_y, rotate_z):
        """Render batches created during class initialization"""


        # Place the light far behind our object
        #z_offset = self.z_min - (self.z_max - self.z_min) * 3
        #light_pos = VecF(100.0, 100.0, 10.0 * -z_offset)
        #glLightfv(GL_LIGHT0, GL_POSITION, (0, 0, 0))
        #glDisable(GL_LIGHT0);
        glPushMatrix();
        #print "push";
        
        # Move the object deeper to the screen and rotate
        glTranslatef(x, y, z);
        glRotatef(rotate_x, 1.0, 0.0, 0.0);
        glRotatef(rotate_y, 0.0, 1.0, 0.0);
        glRotatef(rotate_z, 0.0, 0.0, 1.0);
        
        if self.gllist == None:
            self.gllist = glGenLists(1);
            glNewList(self.gllist, GL_COMPILE);
            prev_shader_prog = None
            # Draw batches (VBOs)
            for (batch, shader_prog, tex_id, diff_color, spec_color, amb_color, shininess) in self.batch_list:
                # Optimization to not make unnecessary bind/unbind for the
                # shader. Most of the time there will be same shaders for
                # geometries.
                if shader_prog != prev_shader_prog:
                    if prev_shader_prog is not None:
                        prev_shader_prog.unbind()
                    prev_shader_prog = shader_prog
                    shader_prog.bind()

                if diff_color is not None:
                    shader_prog.uniformf('diffuse', *diff_color)
                    print 'diffuse';
                if spec_color is not None:
                    shader_prog.uniformf('specular', *spec_color)
                    print 'specular';
                if amb_color is not None:
                    shader_prog.uniformf('ambient', *amb_color)
                    print 'ambience';
                if shininess is not None:
                    shader_prog.uniformf('shininess', shininess)
                    print 'shininess';

                if tex_id is not None:
                    # We assume that the shader here is 'texture'
                    glActiveTexture(GL_TEXTURE0)
                    glEnable(GL_TEXTURE_2D)
                    glBindTexture(GL_TEXTURE_2D, tex_id)
                    shader_prog.uniformi('my_color_texture[0]', 0)

                glBegin(batch[1]);
                
                for i in batch[3]:
                    v = batch[4][1][i*3:i*3+3];
                    n = batch[5][1][i*3:i*3+3];
                    try:
                        t = batch[6][1][i*2:i*2+2];
                        glTexCoord(*t);
                    except:
                        pass;
                    glNormal(*n);
                    glVertex(v);
                
                glEnd();
            if prev_shader_prog is not None:
                prev_shader_prog.unbind()
                
            glEndList();
        else:
            glCallList(self.gllist);
            #print "here";
        
        glPopMatrix();
        #print "pop";
        
    def cleanup(self):
        print 'Renderer cleaning up'
