'''Game main module.

Contains the entry point used by the run_game.py script.

Feel free to put all your game code here, or in other modules in this "gamelib"
package.
'''

import data;

import allegro;
import sys;

from OpenGL.GL import *;
from OpenGL.GLU import *;

from GLSLRenderer import GLSLRenderer;

import collada;

import menu;

from glutils import VecF;

import hexagons;
import environment;
import objects;

import player;
import draw;

import math;

def init_allegro():
    if not allegro.al_install_system(allegro.ALLEGRO_VERSION_INT, None):
        print("Failed to init Allegro.\n");
        sys.exit(1);

    if not allegro.al_init_image_addon():
        print("Failed to init image addon.\n");
        sys.exit(1);
        
    if not allegro.al_init_font_addon() or not allegro.al_init_ttf_addon():
        print("Failed to init font addon.\n");
        sys.exit(1);
        
    if not allegro.al_install_audio():
        print("Failed to init audio addon.\n");
        sys.exit(1);

    allegro.al_init_font_addon();
    
    config = allegro.al_load_config_file(data.filepath("config.txt"));
    if not bool(config):
        config = allegro.al_create_config();
        
        # graphics
        allegro.al_set_config_value(config, "graphics", "width", "1024");
        allegro.al_set_config_value(config, "graphics", "height", "768");
        allegro.al_set_config_value(config, "graphics", "fullscreen", "false");
        allegro.al_set_config_value(config, "graphics", "refresh_rate", "60");
        
        # keys
        allegro.al_set_config_value(config, "keys", "moving_forward", str(allegro.ALLEGRO_KEY_W));
        allegro.al_set_config_value(config, "keys", "moving_backward", str(allegro.ALLEGRO_KEY_S));        
        allegro.al_set_config_value(config, "keys", "turning_left", str(allegro.ALLEGRO_KEY_A));
        allegro.al_set_config_value(config, "keys", "turning_right", str(allegro.ALLEGRO_KEY_D));
        allegro.al_set_config_value(config, "keys", "attacking", str(allegro.ALLEGRO_KEY_LCTRL));
        allegro.al_set_config_value(config, "keys", "defending", str(allegro.ALLEGRO_KEY_LSHIFT));
        allegro.al_set_config_value(config, "keys", "jumping", str(allegro.ALLEGRO_KEY_SPACE));
        allegro.al_set_config_value(config, "keys", "pausing", str(allegro.ALLEGRO_KEY_P));
    
    print "You have", allegro.al_get_num_video_adapters(), "video adapters";
    
    info = allegro.ALLEGRO_MONITOR_INFO();
    allegro.al_get_monitor_info(0, allegro.byref(info));
    w = info.x2 - info.x1;
    h = info.y2 - info.y1;
        
    w = int(allegro.al_get_config_value(config, "graphics", "width"));
    h = int(allegro.al_get_config_value(config, "graphics", "height"));
    print w, h;
    flags = 0;
    
    if allegro.al_get_config_value(config, "graphics", "fullscreen") == "true":
        flags = flags | allegro.ALLEGRO_FULLSCREEN;
    else:
        flags = flags | allegro.ALLEGRO_WINDOWED;
    
    allegro.al_set_new_display_refresh_rate(int(allegro.al_get_config_value(config, "graphics", "refresh_rate")));
    
    flags = flags | allegro.ALLEGRO_OPENGL;
    allegro.al_set_new_display_flags(flags);
    allegro.al_set_new_display_option(allegro.ALLEGRO_DEPTH_SIZE, 16, allegro.ALLEGRO_SUGGEST);
    display = allegro.al_create_display(w, h);
    allegro.al_set_window_title(display, "[Allegro 5 - the beast-hunters]");
    
    if not display:
        print("Error installing keyboard.\n")
        sys.exit(1);

    if not allegro.al_install_keyboard():
        print("Error installing keyboard.\n")
        sys.exit(1);

    if not allegro.al_install_mouse():
        print("Error installing mouse.\n")
        sys.exit(1);
    
    return (display, config);
        
def allegro_init_events(display, timer):
    

    queue = allegro.al_create_event_queue()
    allegro.al_register_event_source(queue, allegro.al_get_keyboard_event_source())
    allegro.al_register_event_source(queue, allegro.al_get_mouse_event_source())
    allegro.al_register_event_source(queue, allegro.al_get_timer_event_source(timer))
    allegro.al_register_event_source(queue, allegro.al_get_display_event_source(display))
    
    return queue;

def main():
    display, config = init_allegro();
    
    timer = allegro.al_create_timer(1.0 / 60.0)
    queue = allegro_init_events(display, timer);
    allegro.al_start_timer(timer)

    game_menu = menu.Menu(config);
    game_menu.show(display, queue);
    
    # hack to stop the title screen showing, when the
    # main menu is diabled.
    game_menu.current_menu = game_menu.main_menu;
    game_menu.display = display;
    game_menu.hide();
    
    pause_key = int(allegro.al_get_config_value(config, "keys", "pausing"));


    frames = 0;
    prev_time = allegro.al_get_time();

    # open COLLADA file ignoring some errors in case they appear
    #collada_file = collada.Collada(data.filepath("cockpit.zip"), 
    #        ignore=[collada.DaeUnsupportedError, collada.DaeBrokenRefError])
    #renderer = GLSLRenderer(collada_file);

    glViewport(0, 0, allegro.al_get_display_width(display), allegro.al_get_display_height(display))
    glMatrixMode(GL_PROJECTION)
    glLoadIdentity()
    gluPerspective(60., float(allegro.al_get_display_width(display)) / float(allegro.al_get_display_height(display)), .1, 1000.)
    glMatrixMode(GL_MODELVIEW)
    glLoadIdentity()
    #glTranslatef(0.0, 0.0, -20.0);
    #glRotatef(0, 1.0, 0.0, 0.0)
    #glRotatef(0, 0.0, 1.0, 0.0)
    #glRotatef(0, 0.0, 0.0, 1.0)


    # massive hack -- this block of code is my RESTART code
    hexagon_grid = environment.environment();
    starting_hexagon = hexagon_grid.get(12, 12*3);
    avatar1 = objects.Character(starting_hexagon.x+0.1, starting_hexagon.y+0.1, 0., starting_hexagon);
    avatar1.hitpoints = 30.;
    avatar1.damaging_objects = [];
    starting_hexagon.objects.append(avatar1); 
    player1 = player.HumanPlayer(avatar1);
    player1.keys_from(config);
    
    beast = objects.Beast(60., 0., 0., hexagon_grid.get(1, 0));
    hexagon_grid.get(1, 0).objects.append(beast);
    beast_player = player.BeastAIPlayer(beast, player1);
    
    light_pos = VecF(0.0, 0.0, 10.0);
    glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
    glLightfv(GL_LIGHT0, GL_AMBIENT, VecF(0.1, 0.1, 0.1, 1.0));
    glLightfv(GL_LIGHT0, GL_DIFFUSE, VecF(0.1, 0.1, 0.1, 1.0));
    glLightfv(GL_LIGHT0, GL_SPECULAR, VecF(0.3, 0.3, 0.3, 1.0));

    damaged_objects = [];
    update_list = [];
    player_list = [];
    paused = False;
    # --------------------------------------------------------
    


    
    glEnable(GL_DEPTH_TEST);
    
    allegro.al_clear_to_color(allegro.al_map_rgb_f(0.1, 0.1, 0.1));


    while True:
        event = allegro.ALLEGRO_EVENT()

        if allegro.al_is_event_queue_empty(queue):
            
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
            glLoadIdentity()
            
            # set the camera to be the same position as 
            # player1's avatar
            gluLookAt(avatar1.x, avatar1.y, 0., avatar1.x, avatar1.y, -20., 0.0, 1.0, 0.0);
            
            # Place the light far behind our object
            #light_pos = VecF(100.0, 100.0, 10.0 * -20)
            #glLightfv(GL_LIGHT0, GL_POSITION, light_pos)
            
            
            #renderer.render(0, 0, -20, 0, 0, 0);
            
            
            draw.draw_environment(hexagon_grid, avatar1.hexagon);
            
            #draw.draw_character(avatar1);
            #draw.draw_beast(beast);
            
            
            if avatar1.hexagon.ring == 12 and avatar1.hexagon.segment == 12*3:
                draw.draw_right_arrow(avatar1);
                draw.draw_backstory(display,
                [
                    "For generations your family has\n hunted the beast of the forest\n"+
                    " as an initiation ritual.",
                    "It is widely regarded as the hardest\n"
                    " initiation ritual in the region, with\n"
                    " no relative ever being successfully\n"+
                    " initiated",
                    "Now it is your turn to be initiated.\n"+
                    "Follow the arrow."
                ]);
            elif avatar1.hexagon.ring == 11 and avatar1.hexagon.segment == 11*3:
                draw.draw_right_arrow(avatar1);
                draw.draw_backstory(display,
                [
                    "The forest floor is covered in\n"+
                    "minions of the beast\n"+
                    "Prepare to fight.",
                    
                    "Fighting is best achieved by\n"+
                    " facing your sword\n"+
                    " (Yes the red line...)\n"+
                    " toward the enemies and\n"+
                    " tapping the attacking button."
                ]);
            elif avatar1.hexagon.ring == 10 and avatar1.hexagon.segment == 10*3:
                if beast.hitpoints < 0:
                    draw.draw_backstory(display,
                    [
                        "Congratulations!",
                        
                        "You have succeeded where\n",
                        "Others in your family have failed."
                    ]);
                    paused = True;
                else:
                    draw.draw_right_arrow(avatar1);
                    draw.draw_line_between_beast_and_player(beast, avatar1);
                    draw.draw_backstory(display,
                    [
                        "Every member of your family can\n"+
                        " sense the beast",
                        
                        "In order to find the beast\n"+
                        " follow the green line and enter\n"+
                        " the forrest...",
                        
                        "The beast is your nemesis"
                    ]);
            elif beast.hitpoints < 0:
                draw.draw_backstory(display,
                [
                    "Congratulations... now you just have\n"+
                    "to make it out alive..."
                ]);
            elif avatar1.hitpoints < 0:
                draw.draw_backstory(display,
                [
                    "GAME OVER",
                    "Well, you followed in family tradition"
                ]);
                paused = True;
                    
            else:
                draw.draw_line_between_beast_and_player(beast, avatar1);

            allegro.al_flip_display();
            frames += 1;
        

        allegro.al_wait_for_event(queue, allegro.byref(event));

        if event.type == allegro.ALLEGRO_EVENT_TIMER:
            
            if not paused:
                
                player1.update();
                
                if avatar1.sword.state == "reset":
                    damaged_objects = [];
                    
                avatar1.update(hexagon_grid);
                
                if not avatar1.hexagon.is_point_inside(avatar1.x, avatar1.y):
                    # move the avatar back inside
                    point = avatar1.hexagon.get_closest_point(avatar1.x, avatar1.y);
                    #point = (point[0] + avatar1.hexagon.x, point[1] + avatar1.hexagon.y);
                    old_point = (avatar1.x, avatar1.y);
                    avatar1.x, avatar1.y = point;
                    #draw.draw_tree(avatar1);
                    #avatar1.x, avatar1.y = old_point


                beast_player.update();
                beast.update(hexagon_grid);

                
                if avatar1.sword.state == "swinging" and avatar1.sword.omega > 0.:
                    for hexagon in [avatar1.hexagon] + hexagon_grid.neighbours(avatar1.hexagon):
                        for obj in hexagon.objects:
                            if obj != avatar1 and avatar1.sword.collision_with(obj) and obj not in damaged_objects:
                                damaged_objects.append(obj);
                                if obj.type == "minion" and obj.sword.state == "swinging" and obj.sword.omega > 0. and\
                                    obj.sword.collision_with(avatar1):
                                    # knock back
                                    #print "knock_back"
                                    angle_betw = math.atan2(avatar1.y - obj.y, avatar1.x - obj.x);
                                    avatar1.x += 1. * math.cos(angle_betw);
                                    avatar1.y += 1. * math.sin(angle_betw);
                                    obj.x += 1. * math.cos(angle_betw+math.pi);
                                    obj.y += 1. * math.sin(angle_betw+math.pi);
                                else:
                                    obj.take_damage(avatar1.sword.damage);
            
                #print avatar1.hexagon;
                #print avatar1.hexagon.ring;
                #print avatar1.hexagon.objects;
            
                for hexagon in [avatar1.hexagon]:# + hexagon_grid.neighbours(avatar1.hexagon):
                    for obj in hexagon.objects:
                        if obj.type == "minion" and obj not in update_list:
                            update_list.append(obj);
                            player_list.append(player.MinionAIPlayer(obj, player1));
                         
                        for obj2 in hexagon.objects:
                            dist_betw = math.sqrt((obj.x - obj2.x)**2. + (obj.y - obj2.y)**2.);
                            
                            if dist_betw < obj.radius + obj2.radius:
                                if obj != obj2 and obj.type != "tree" and obj2.type == "tree":
                                    # push away
                                    #print dist_betw;
                                    #print obj.radius + obj2.radius;
                                    angle_betw = math.atan2(obj.y - obj2.y, obj.x - obj2.x);
                                    obj.x += (obj.radius+obj2.radius - dist_betw) * math.cos(angle_betw);
                                    obj.y += (obj.radius+obj2.radius - dist_betw) * math.sin(angle_betw);
                                    #print "push away";
                
                for minion_player in player_list:
                    minion_player.update();

                for obj in update_list:
                    if obj not in obj.hexagon.objects:
						# likely dead
                        update_list.remove(obj);
                    
                    if obj.type == "minion":
                        if obj.sword.state == "reset":
                            if obj in avatar1.damaging_objects:
                                avatar1.damaging_objects.remove(obj);
                    
                    obj.update(hexagon_grid);
                    if obj.type == "minion":
                        if obj.sword.state == "swinging" and obj.sword.omega > 0.:
                            if obj.sword.collision_with(avatar1):
                                avatar1.damaging_objects.append(obj);
                                if avatar1.sword.state == "swinging" and avatar1.sword.omega > 0. and\
                                    avatar1.sword.collision_with(obj):
                                    # knock back
                                    #print "knock back"
                                    angle_betw = math.atan2(avatar1.y - obj.y, avatar1.x - obj.x);
                                    avatar1.x += 1. * math.cos(angle_betw);
                                    avatar1.y += 1. * math.sin(angle_betw);
                                    obj.x += 1. * math.cos(angle_betw+math.pi);
                                    obj.y += 1. * math.sin(angle_betw+math.pi);
                                else:
                                    avatar1.take_damage(obj.sword.damage);
                                    
                for hexagon in [beast.hexagon] + hexagon_grid.neighbours(beast.hexagon):
                    for obj in hexagon.objects:
                        if obj != beast and obj.type == "tree":
                            dist_betw = math.sqrt((obj.x - beast.x)**2. + (obj.y - beast.y)**2.);
                            if dist_betw < obj.radius + beast.radius:
                                obj.take_damage(100.);


        
        elif event.type == allegro.ALLEGRO_EVENT_DISPLAY_CLOSE:
            break;
            
        elif event.type == allegro.ALLEGRO_EVENT_KEY_UP or event.type == allegro.ALLEGRO_EVENT_KEY_DOWN:
            if event.keyboard.keycode == allegro.ALLEGRO_KEY_ESCAPE and event.type == allegro.ALLEGRO_EVENT_KEY_DOWN:
                game_menu.show(display, queue);
                player1.keys_from(game_menu.config);
                pause_key = int(allegro.al_get_config_value(config, "keys", "pausing"));
                if game_menu.restart:
                    # massive hack -- this block of code is my RESTART code
                    hexagon_grid = environment.environment();
                    starting_hexagon = hexagon_grid.get(12, 12*3);
                    avatar1 = objects.Character(starting_hexagon.x+0.1, starting_hexagon.y+0.1, 0., starting_hexagon);
                    avatar1.hitpoints = 30.;
                    avatar1.damaging_objects = [];
                    starting_hexagon.objects.append(avatar1); 
                    player1 = player.HumanPlayer(avatar1);
                    player1.keys_from(config);
                    
                    beast = objects.Beast(60., 0., 0., hexagon_grid.get(1, 0));
                    hexagon_grid.get(1, 0).objects.append(beast);
                    beast_player = player.BeastAIPlayer(beast, player1);
                    
                    light_pos = VecF(0.0, 0.0, 10.0);
                    glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
                    glLightfv(GL_LIGHT0, GL_AMBIENT, VecF(0.1, 0.1, 0.1, 1.0));
                    glLightfv(GL_LIGHT0, GL_DIFFUSE, VecF(0.1, 0.1, 0.1, 1.0));
                    glLightfv(GL_LIGHT0, GL_SPECULAR, VecF(0.3, 0.3, 0.3, 1.0));

                    damaged_objects = [];
                    update_list = [];
                    player_list = [];
                    paused = False;
                    # --------------------------------------------------------
                    game_menu.restart = False;
            elif event.keyboard.keycode == pause_key and event.type == allegro.ALLEGRO_EVENT_KEY_DOWN:
                paused = not paused;
                #print "paused", paused;
            else:
                player1.on_event(event);

        current_time = allegro.al_get_time();   
        if current_time - prev_time > 10.0:
            fps = frames / (current_time - prev_time);
            allegro.al_set_window_title(display, "[Allegro 5 - the beast-hunters] - FPS: %f" % fps);         
            frames = 0;
            prev_time = current_time;
        
    
