
import math;
import utils;

class GameObject:
    def __init__(self, type, x, y, facing, radius, hexagon, hitpoints = 1.):
        self.x = x;
        self.y = y;
        self.facing = facing;
        self.speed_x = 0;
        self.speed_y = 0;
        self.omega = 0;
        self.hexagon = hexagon;
        self.type = type;
        self.radius = radius;
        self.hitpoints = hitpoints;
        
    def set_speed(self, amt):
        self.speed_x = amt * math.cos(self.facing);
        self.speed_y = amt * math.sin(self.facing);
        
    def set_omega(self, amt):
        self.omega = amt;
        
    def update(self, environment):
        self.x += self.speed_x;
        self.y += self.speed_y;
        self.facing += self.omega;
        
        if self.facing > math.pi:
            self.facing -= 2. * math.pi;
        elif self.facing < -math.pi:
            self.facing += 2. * math.pi;
            
        for hexagon in [self.hexagon] + environment.neighbours(self.hexagon):
                if hexagon.is_point_inside(self.x, self.y):
                    if hexagon != self.hexagon:
                        self.hexagon.objects.remove(self);
                        self.hexagon = hexagon;
                        hexagon.objects.append(self);
                    break;
    
    def take_damage(self, amt):
        if self.hitpoints < 0: return;
        self.hitpoints -= amt;
        if self.hitpoints < 0:
            self.hexagon.objects.remove(self);
        
class Sword:
    def __init__(self, character, x_offset, y_offset, initial_facing, final_facing, length, damage):
        self.character = character;
        self.initial_facing = initial_facing;
        self.final_facing = final_facing;
        self.facing = initial_facing;
        self.state = "swinging";
        self.omega = 0;
        self.length = length;
        self.x = x_offset;
        self.y = y_offset;
        self.damage = damage;
        
    def reset(self):
        self.facing = self.initial_facing;
        self.state = "reset";
        
    def set_omega(self, omega):
        self.omega = omega;
        
    def base(self):
        return (self.x*math.cos(self.character.facing) - self.y*math.sin(self.character.facing),
            self.x * math.sin(self.character.facing) + self.y * math.cos(self.character.facing));
        
    def tip(self):
        if self.state == "swinging":
            x, y = self.base();
            return (x + self.length * math.cos(self.facing + self.character.facing), y + self.length * math.sin(self.facing+self.character.facing));
        else:
            return self.base();
        
    def update(self):
        if self.state == "swinging":
            self.facing += self.omega;
            if self.facing >= self.final_facing:
                self.state = "resting";
                self.rest_count = 60;
                
        elif self.state == "resting":
            self.rest_count -= 1;
            if self.rest_count <= 0:
                self.reset();
                
        elif self.state == "reset":
            self.state = "swinging";
                
    def collision_with(self, obj):
        base = self.base();
        x = base[0] + self.character.x;
        y = base[1] + self.character.y;
        
        dist_betw = math.sqrt((obj.x - x)**2. + (obj.y - y)**2.);
        angle_betw = math.atan2(obj.y -  y, obj.x - x);
        
        if dist_betw < self.length + obj.radius:
            if utils.compare_angles(angle_betw, self.facing + self.character.facing) < 0 and\
                utils.compare_angles(angle_betw, self.character.facing + self.initial_facing) > 0:
                return True;
        return False;
            
class Character(GameObject):
    def __init__(self, x, y, facing, hexagon, type = None):
        if type == None:
            type = "character";
        
        GameObject.__init__(self, type, x, y, facing, 1., hexagon);
        self.attacking = False;
        self.defending = False;
        self.jumping = False;
        self.sword = Sword(self, 1., -2., -65.*math.pi/180., 65.*math.pi/180., 6., 1.);
    
    def set_attacking(self, attacking):
        if attacking and not self.attacking:
            self.sword.set_omega(0.2);
        elif self.attacking and not attacking:
            self.sword.reset();
            self.sword.set_omega(0.);
        self.attacking = attacking; 

    def set_defending(self, defending):
        self.defending = defending;
        
    def set_jumping(self, jumping):
        self.jumping = jumping;
    
    def update(self, environment):
        GameObject.update(self, environment);
        self.sword.update();
        
class Minion(Character):
    def __init__(self, x, y, facing, hexagon):
        Character.__init__(self, x, y, facing, hexagon, "minion");

class Beast(Character):
    def __init__(self, x, y, facing, hexagon):
        Character.__init__(self, x, y, facing, hexagon, "beast");
        self.hitpoints = 200.;
        self.radius = 3.;
